/*
 * NBCT
 * Copyright (c) 2007-2009, Yoshitaka Kameya
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,  with or without
 * modification,  are permitted provided that the following conditions
 * are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions  in  binary  form  must  reproduce  the  above
 *    copyright notice,  this list of conditions  and  the following
 *    disclaimer  in  the  documentation  and/or   other   materials
 *    provided with the distribution.
 *  * Neither  the  name  of  Tokyo Institute of Technology  nor the
 *    names of its contributors may be used to  endorse  or  promote
 *    products derived from  this  software  without  specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY  AND FITNESS
 * FOR  A  PARTICULAR  PURPOSE  ARE  DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,    SPECIAL,   EXEMPLARY,   OR   CONSEQUENTIAL   DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS;  OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR  TORT  (INCLUDING  NEGLIGENCE  OR OTHERWISE)
 * ARISING  IN  ANY  WAY  OUT  OF  THE  USE  OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "nbc_def.h"
#include "nbc_ext.h"

#define DESC_ID_STR_SIZE 128
#define DESC_COL_STR_SIZE 128
#define DESC_COL_SIZE 10
#define DESC_COMM_STR_SIZE 1024
#define DESC_COMM_SIZE 3
#define DATA_COL_SIZE 20
#define DATA_COL_STR_SIZE (MAX_WORD_LEN + 1)

static char desc_id[DESC_ID_STR_SIZE];
static char **desc_col;
static char **desc_comm;
static char **data_col;
static int desc_col_size;
static int desc_comm_size;
static int data_col_size;
static int max_desc_col_size;
static int max_desc_comm_size;
static int max_data_col_size;

static void initialize_desc_col(void)
{
  int i;

  desc_col = (char **)xmalloc(sizeof(char *)*DESC_COL_SIZE);
  for (i=0; i<DESC_COL_SIZE; i++)
    desc_col[i] = alloc_str(DESC_COL_STR_SIZE);

  desc_col_size = 0;
  max_desc_col_size = DESC_COL_SIZE;
}

static void expand_desc_col(void)
{
  int i,old_size,new_size;

  old_size = max_desc_col_size;
  new_size = old_size * 2;

  desc_col = (char **)xrealloc(desc_col,sizeof(char *)*new_size);
  for (i=old_size; i<new_size; i++)
    desc_col[i] = alloc_str(DESC_COL_STR_SIZE);

  max_desc_col_size = new_size;
}

static void release_desc_col(void)
{
  int i;
  for (i=0; i<max_desc_col_size; i++)
    xfree(desc_col[i]);
  xfree(desc_col);
}

static void initialize_desc_comm(void)
{
  int i;

  desc_comm = (char **)xmalloc(sizeof(char *)*DESC_COMM_SIZE);
  for (i=0; i<DESC_COMM_SIZE; i++)
    desc_comm[i] = alloc_str(DESC_COMM_STR_SIZE);

  desc_comm_size = 0;
  max_desc_comm_size = DESC_COMM_SIZE;
}

static void expand_desc_comm(void)
{
  int i,old_size,new_size;

  old_size = max_desc_comm_size;
  new_size = old_size * 2;

  desc_comm = (char **)xrealloc(desc_comm,sizeof(char *)*new_size);
  for (i=old_size; i<new_size; i++)
    desc_comm[i] = alloc_str(DESC_COMM_STR_SIZE);

  max_desc_comm_size = new_size;
}

static void release_desc_comm(void)
{
  int i;
  for (i=0; i<max_desc_comm_size; i++)
    xfree(desc_comm[i]);
  xfree(desc_comm);
}

static void initialize_data_col(void)
{
  int i;

  data_col = (char **)xmalloc(sizeof(char *)*DATA_COL_SIZE);
  for (i=0; i<DATA_COL_SIZE; i++)
    data_col[i] = alloc_str(DATA_COL_STR_SIZE);

  data_col_size = 0;
  max_data_col_size = DATA_COL_SIZE;
}

static void expand_data_col(void)
{
  int i,old_size,new_size;

  old_size = max_data_col_size;
  new_size = old_size * 2;

  data_col = (char **)xrealloc(data_col,sizeof(char *)*new_size);
  for (i=old_size; i<new_size; i++)
    data_col[i] = alloc_str(DATA_COL_STR_SIZE);

  max_data_col_size = new_size;
}

static void release_data_col(void)
{
  int i;
  for (i=0; i<max_data_col_size; i++)
    xfree(data_col[i]);
  xfree(data_col);
}

void initialize_printer(void)
{
  initialize_desc_col();
  initialize_desc_comm();
  initialize_data_col();
}

void release_printer(void)
{
  release_desc_col();
  release_desc_comm();
  release_data_col();
}

void print_usage(FILE *fp)
{
  fprintf(fp,"%s",conv_msg(usage));
}

void print_usage_full(FILE *fp)
{
  fprintf(fp,"%s",conv_msg(usage_full));
}

void add_desc_id(char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
#ifdef _MSC_VER
  _vsnprintf(desc_id,(DESC_ID_STR_SIZE - 1),conv_desc(fmt),ap);
  desc_id[DESC_ID_STR_SIZE - 1] = '\0';
#else
  vsnprintf(desc_id,DESC_ID_STR_SIZE,conv_desc(fmt),ap);
#endif
  va_end(ap);
}

static void clean_desc(void)
{
  int i;

  memset(desc_id,0,DESC_ID_STR_SIZE);
  for (i=0; i<desc_col_size; i++)
    memset(desc_col[i],0,sizeof(char)*DESC_COL_STR_SIZE);
  for (i=0; i<desc_comm_size; i++)
    memset(desc_comm[i],0,sizeof(char)*DESC_COMM_STR_SIZE);
  desc_col_size = 0;
  desc_comm_size = 0;
}

static void clean_data(void)
{
  int i;

  for (i=0; i<data_col_size; i++)
    memset(data_col[i],0,sizeof(char)*DATA_COL_STR_SIZE);
  data_col_size = 0;
}

void flush_desc(FILE *fp)
{
  int i;

  fprintf(fp,conv_crlf("#### ID:%s\n"),desc_id);

  if (desc_col_size > 0) {
    fprintf(fp,"#### ");
    for (i=0; i<desc_col_size; i++) {
      if (i > 0) fprintf(fp,", ");
      fprintf(fp,"%s",desc_col[i]);
    }
    fprintf(fp,"%s",conv_crlf("\n"));
  }

  if (desc_comm_size == 0) {
    fprintf(fp,"%s",conv_crlf("####\n"));
  }
  else {
    for (i=0; i<desc_comm_size; i++)
      fprintf(fp,conv_crlf("#### -- %s\n"),desc_comm[i]);
    fprintf(fp,"%s",conv_crlf("####\n"));
  }

  clean_desc();

  fflush(fp);
}

void flush_data(FILE *fp)
{
  int i;

  if (data_col_size <= 0) return;

  for (i=0; i<data_col_size; i++) {
    if (i > 0) fprintf(fp,"%c",((format_out == FMT_CSV) ? ',' : '\t'));
    fprintf(fp,"%s",data_col[i]);
  }
  fprintf(fp,"%s",conv_crlf("\n"));
  clean_data();

  fflush(fp);
}

void print_data(FILE *fp, char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp,conv_data(fmt),ap);
  va_end(ap);

  fprintf(fp,"%s",conv_crlf("\n"));
}

void add_desc_col(char *fmt, ...)
{
  va_list ap;
  char *s;

  if (desc_col_size == max_desc_col_size) expand_desc_col();
  s = desc_col[desc_col_size];

  va_start(ap,fmt);
#ifdef _MSC_VER
  _vsnprintf(s,DESC_COL_STR_SIZE,conv_desc(fmt),ap);
  s[DESC_COL_STR_SIZE - 1] = '\0';
#else
  vsnprintf(s,DESC_COL_STR_SIZE,conv_desc(fmt),ap);
#endif
  va_end(ap);

  desc_col_size++;
}

void add_desc_comm(char *fmt, ...)
{
  va_list ap;
  char *s;

  if (desc_comm_size == max_desc_comm_size) expand_desc_comm();
  s = desc_comm[desc_comm_size];

  va_start(ap,fmt);
#ifdef _MSC_VER
  _vsnprintf(s,DESC_COMM_STR_SIZE,conv_desc(fmt),ap);
  s[DESC_COMM_STR_SIZE - 1] = '\0';
#else
  vsnprintf(s,DESC_COMM_STR_SIZE,conv_desc(fmt),ap);
#endif
  va_end(ap);

  desc_comm_size++;
}

void add_data(char *fmt, ...)
{
  va_list ap;
  char *s;

  if (data_col_size == max_data_col_size) expand_data_col();
  s = data_col[data_col_size];

  va_start(ap,fmt);
#ifdef _MSC_VER
  _vsnprintf(s,DATA_COL_STR_SIZE,conv_data(fmt),ap);
  s[DATA_COL_STR_SIZE - 1] = '\0';
#else
  vsnprintf(s,DATA_COL_STR_SIZE,conv_data(fmt),ap);
#endif
  va_end(ap);

  data_col_size++;
}

/* [Note]
 * This function returns a string allocated in the static area.
 * So we should make a copy of the returned string if necessary
 * (e.g. using together with add_data()).
 */
char *get_case_id(int e)
{
  static char s[INT2STR_BUFSIZE];

  if (mode_id)
    return case_id_list[freq_list[e].case_id];
  else {
    strncpy(s,int2str(e),INT2STR_BUFSIZE);
    return s;
  }
}

char *get_ans_class(int e)
{
  return ans_class_list[freq_list[e].ans_id];
}

/* [Note]
 * This function returns a string allocated in the static area.
 * So we should make a copy of the returned string if necessary
 * (e.g. using together with add_data()).
 */
char *freq_entry_val(int e, int v)
{
  static char *c = NULL;
  char *p = NULL;
  int u,a,v_d;

  if (is_attr_cont(v)) {
    if (c == NULL) c = alloc_str(MAX_FLOAT_LEN);

    u = attr_info[v].gauss_id;
    a = attr_info[v].local_id;

    if (freq_list[e].miss_g[u][a])
      conv_sprintf(c,"?");
    else
      conv_snprintf(c,MAX_FLOAT_LEN,"%.Ng",freq_list[e].gval[u][a]);

    p = c;
  }
  else {
    v_d = attr_info[v].local_id;

    if (freq_list[e].id[v_d] >= 0)
      p = word_list[v_d][freq_list[e].id[v_d]];
    else if (freq_list[e].miss_w[v_d])
      p = "?";
    else die("Internal error: freq_entry_val(): word not assigned");
  }

  return p;
}

char *mode_str(int mode)
{
  return (mode > 0) ? "Enabled" : "Disabled";
}
